@@ -44,6 +44,11 @@
 #define LINUX_CPIO_UCODE_NAME "kernel/x86/microcode/GenuineIntel.bin"
 #define LINUX_CPIO_FILE_MODE 0100644
 #define LINUX_CPIO_DIR_MODE  0040755
+#define LINUX_CPIO_BLK_SIZE  (1024U)
+
+/* file load constants */
+#define IUCODE_TYPICAL_MCU_FILE_SIZE (0x100000UL) /* 1MiB */
+#define IUCODE_MAX_MCU_FILE_SIZE  (0x40000000ULL) /* 1GiB, arbitrary */
 
 /* Command-line processing and UI */
 
@@ -113,9 +118,14 @@ static unsigned long int last_bundle_id = 0;
 
 /* intel_uclist_entry signature flag constants */
 enum {
-    INTEL_UCLE_EXTSIG = 0x0001,	/* Sig came from extended table */
-    INTEL_UCLE_SELID  = 0x0010,	/* Sig is candidate for loose datefilter */
-    INTEL_UCLE_SELECT = 0x0020,	/* Sig selected */
+    INTEL_UCLE_EXTSIG = 0x0001, /* Sig came from extended sig table */
+    INTEL_UCLE_HASXST = 0x0002, /* ucode has an extended sig table */
+    INTEL_UCLE_SELID  = 0x0010, /* Sig is candidate for loose datefilter */
+    INTEL_UCLE_SELECT = 0x0020, /* Sig selected */
+    INTEL_UCLE_NOWR   = 0x0100, /* Do not write this sig to output
+                                 *  + When INTEL_UCLE_HASXST is set,
+                                 *    INTEL_UCLE_NOWR is reserved for
+                                 *    uclist_annotate_extsig_dup() */
 };
 
 /* signature single-linked list */
@@ -165,6 +175,13 @@ static inline int filter_list_active(void)
 		!!(command_line_actions & IUCODE_F_UCSELECT);
 }
 
+/* extended sig dedup */
+struct microcode_id_entry {
+	const void * id;	/* exact match */
+	struct microcode_id_entry *next;
+};
+static unsigned int extsig_tables_in_use = 0;
+
 /* Helpers */
 
 #define print_msg_u(format, arg...) \
@@ -328,6 +345,7 @@ static int uclist_add_signature(const unsigned long int id,
  * uclist_merge_signature() - merge signature into ucode sig list
  * @id:		microcode id
  * @gid:	microcode group id
+ * @flags:	INTEL_UCLE_* flags
  * @cpuid:	signature of the signature entry being added
  * @pf_mask:	pf_mask of the signature entry being added
  * @uc_rev:	revision of the microcode entry being added
@@ -357,6 +375,7 @@ static int uclist_add_signature(const unsigned long int id,
  */
 static int uclist_merge_signature(const unsigned long int id,
 			const unsigned long int gid,
+			const uint32_t flags,
 			const uint32_t cpuid,
 			const uint32_t pf_mask,
 			const int32_t uc_rev,
@@ -395,6 +414,7 @@ static int uclist_merge_signature(const unsigned long int id,
 					 * one is newer microcode: replace */
 					e->id = id;
 					e->gid = gid;
+					e->flags = flags;
 					/* e->cpuid = cpuid; */
 					e->pf_mask = pf_mask;
 					e->uc_rev = uc_rev;
@@ -409,6 +429,7 @@ static int uclist_merge_signature(const unsigned long int id,
 					/* we can replace the old entry */
 					e->id = id;
 					e->gid = gid;
+					e->flags = flags;
 					/* e->cpuid = cpuid; */
 					e->pf_mask = pf_mask;
 					e->uc_rev = uc_rev;
@@ -428,6 +449,7 @@ static int uclist_merge_signature(const unsigned long int id,
 				/* same set, replace */
 				e->id = id;
 				e->gid = gid;
+				e->flags = flags;
 				/* e->cpuid = cpuid; */
 				e->pf_mask = pf_mask;
 				e->uc_rev = uc_rev;
@@ -448,6 +470,7 @@ static int uclist_merge_signature(const unsigned long int id,
 	memset(n, 0, sizeof(struct intel_uclist_entry));
 	n->id = id;
 	n->gid = gid;
+	n->flags = flags;
 	n->cpuid = cpuid;
 	n->pf_mask = pf_mask;
 	n->uc_rev = uc_rev;
@@ -526,7 +549,8 @@ static void free_intel_microcode_bundles(void)
 static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length)
 {
 	const size_t line_buffer_size = 2048;
-	const size_t buffer_size_granularity = 256*1024; /* 1MiB, dwords */
+	const size_t buffer_size_granularity = IUCODE_TYPICAL_MCU_FILE_SIZE / sizeof(uint32_t);
+	const size_t buffer_limit = IUCODE_MAX_MCU_FILE_SIZE / sizeof(uint32_t);
 	char *line_buffer = NULL;
 	char *lp;
 	size_t mcb_buflen;
@@ -540,7 +564,7 @@ static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length)
 	assert(mcb_length);
 
 	mcb_buflen = buffer_size_granularity;
-	mcb_buffer = malloc(buffer_size_granularity * sizeof(uint32_t));
+	mcb_buffer = malloc(mcb_buflen * sizeof(uint32_t));
 	line_buffer = malloc(line_buffer_size);
 	if (!mcb_buffer || !line_buffer) {
 		err = ENOMEM;
@@ -585,8 +609,12 @@ static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length)
 
 			/* we did read a value */
 			pos++;
-			if (unlikely(mcb_buflen < pos)) {
+			if (unlikely(mcb_buflen <= pos)) {
 				/* expand buffer */
+				if (unlikely(buffer_limit - mcb_buflen < buffer_size_granularity)) {
+					err = EFBIG;
+					goto err_out;
+				}
 				mcb_buflen += buffer_size_granularity;
 				rp = realloc(mcb_buffer,
 					     mcb_buflen * sizeof(uint32_t));
@@ -641,7 +669,8 @@ err_out:
 
 static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
 {
-	const size_t buffer_size_granularity = 1024*1024;
+	const size_t buffer_size_granularity = IUCODE_TYPICAL_MCU_FILE_SIZE;
+	const size_t buffer_limit = IUCODE_MAX_MCU_FILE_SIZE;
 
 	int file_size_known;
 	struct stat stat;
@@ -659,6 +688,8 @@ static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
 
 	/* Try to get file size beforehand */
 	if (!fstat(fd, &stat) && S_ISREG(stat.st_mode) && stat.st_size > 0) {
+		if ((size_t)stat.st_size > buffer_limit)
+			return EFBIG;
 		mcb_size = stat.st_size;
 		file_size_known = 1;
 	} else {
@@ -701,6 +732,10 @@ static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
 			break; /* Read entire file */
 
 		if (!mcb_space_left) {
+			if (unlikely(buffer_limit - mcb_size < buffer_size_granularity)) {
+				err = EFBIG;
+				goto err_out;
+			}
 			mcb_size += buffer_size_granularity;
 			mcb_space_left += buffer_size_granularity;
 			rp = realloc(mcb_buffer, mcb_size);
@@ -727,12 +762,12 @@ static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
 	*mcb = rp;
 	*mcb_length = mcb_size;
 
+	return 0;
+
 err_out:
-	if (err) {
-		free(mcb_buffer);
-		*mcb = NULL;
-		*mcb_length = 0;
-	}
+	free(mcb_buffer);
+	*mcb = NULL;
+	*mcb_length = 0;
 
 	return err;
 }
@@ -772,7 +807,7 @@ static int load_intel_microcode_file(int fd,
 		goto err_out;
 	}
 
-	if (!err && mcb && mcb_length) {
+	if (!err && mcb) {
 		last_bundle_id++;
 		err = add_intel_microcode_bundle(fn, last_bundle_id,
 						 mcb, mcb_length);
@@ -791,6 +826,9 @@ static int load_intel_microcode_file(int fd,
 	case EINVAL:
 		print_err("%s: invalid file format", fn);
 		break;
+	case EFBIG:
+		print_err("%s: cowardly refusing to load an insanely large data file", fn);
+		break;
 	}
 
 err_out:
@@ -799,6 +837,8 @@ err_out:
 	else if (fd != -1)
 		close(fd);
 
+	free(mcb);
+
 	return err;
 }
 
@@ -836,8 +876,10 @@ static int load_intel_microcode(const char * path,
 
 		err = 0;
 
-		if (fd != -1)
+		if (fd != -1) {
 			close(fd);
+			fd = -1;
+		}
 
 		if (dir) {
 			errno = 0;
@@ -1018,22 +1060,27 @@ static int upload_intel_microcodes(const char * const devname,
 	}
 
 	while (uc_write_list && uc_write_list->uc) {
-		if (verbosity >= 3)
-			log_microcode_action("uploading",
-					     devname, uc_write_list);
+		if (likely(!(uc_write_list->flags & INTEL_UCLE_NOWR))) {
+			if (verbosity >= 3)
+				log_microcode_action("uploading",
+						     devname, uc_write_list);
 
-		rc = write_data(fd, uc_write_list->uc,
-				  uc_write_list->uc_size);
-		if (rc < 0) {
-			err = -rc;
-			print_err("%s: write error: %s",
-				  devname, strerror(err));
-			break;
-		} else {
+			rc = write_data(fd, uc_write_list->uc,
+					  uc_write_list->uc_size);
+			if (rc < 0) {
+				err = -rc;
+				print_err("%s: write error: %s",
+					  devname, strerror(err));
+				break;
+			}
 			total_written += rc;
-			uc_write_list = uc_write_list->next;
 			entries_written++;
+		} else {
+			if (verbosity >= 3)
+				log_microcode_action("skipping",
+						     devname, uc_write_list);
 		}
+		uc_write_list = uc_write_list->next;
 	}
 
 	if (close(fd)) {
@@ -1061,6 +1108,7 @@ static int xx_write_cpio_hdrentry(int fd, const char * const name,
 	char buf[CPIO_MAX_HDRSIZE + 16]; /* worst case padding */
 	int nsize;
 	int bufsize;
+	int s;
 
 	assert(name);
 	nsize = strlen(name) + 1;
@@ -1069,17 +1117,15 @@ static int xx_write_cpio_hdrentry(int fd, const char * const name,
 
 	/* pad to DWORD */
 	bufsize = CPIO_STATIC_HDRSIZE + nsize;
-	while ((*pos + bufsize) % 4)
-		bufsize++;
+	bufsize += (4 - (*pos + bufsize) % 4) % 4;
 
 	/* Gross hack to work around a Linux kernel bug: for file
 	 * entries, force file data into a 16-byte alignment by
 	 * appending NULs to the file name.  Verified to be compatible
 	 * with GNU pax, and GNU cpio */
-	while (size && (*pos + bufsize) % 16) {
-		bufsize++;
-		nsize++;
-	}
+	s = (size) ? (16 - (*pos + bufsize) % 16) % 16 : 0;
+	bufsize += s;
+	nsize += s;
 
 	*pos += bufsize;
 
@@ -1118,6 +1164,12 @@ static int write_cpio_header(int fd, size_t size)
 	size_t pos = 0;
 	int r;
 
+	/*
+	 * the early initramfs kernel loader doesn't need the directories
+	 * to be created before the file can be accessed, but the regular
+	 * initramfs kernel loader does.  Including them makes the file
+	 * available inside the regular initramfs.
+	 */
 	p1 = fn;
 	do {
 		p2 = strchr(p1, '/');
@@ -1147,8 +1199,8 @@ static int write_cpio_header(int fd, size_t size)
  * if a problem happens
  *
  * Write a cpio EOF trailer, padding with NULs to the nearest
- * 16-byte boundary.  @pos is used to calculate the amount of
- * padding, instead of requiring a lseek().
+ * LINUX_CPIO_BLK_SIZE boundary.  @pos is used to calculate the
+ * amount of padding, instead of requiring a lseek().
  */
 static int write_cpio_trailer(int fd, size_t pos)
 {
@@ -1160,15 +1212,24 @@ static int write_cpio_trailer(int fd, size_t pos)
 		/* devm,  rdevj, rdevm, name size */
 		"00000000" "00000000" "00000000" "0000000B"
 		/* checksum, name */
-		"00000000" "TRAILER!!!"
-		/* 16 extra NULs for padding */
-		"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
+		"00000000" "TRAILER!!!";
+	size_t s = sizeof(cpio_trailer) +
+		((LINUX_CPIO_BLK_SIZE -
+		  ((pos + sizeof(cpio_trailer)) % LINUX_CPIO_BLK_SIZE))
+		 % LINUX_CPIO_BLK_SIZE);
+	char *buf;
+	ssize_t rc;
 
-	size_t s = sizeof(cpio_trailer) - 16;
-	while ((pos + s) % 16)
-		s++;
+	buf = malloc(s);
+	if (!buf)
+		return -ENOMEM;
 
-	return write_data(fd, cpio_trailer, s);
+	/* we depend on the buffer-fill strncpy() semanthics here */
+	rc = write_data(fd, strncpy(buf, cpio_trailer, s), s);
+
+	free(buf);
+
+	return rc;
 }
 
 
@@ -1179,7 +1240,10 @@ static int write_cpio_trailer(int fd, size_t pos)
  * @ft:			file type
  * @uc_write_list:	uclist with the microcodes to write
  *
- * returns 0 if successful, or errno() if a problem happens
+ * returns 0 if successful, errno() if a problem happens,
+ * or ENODATA if all microcodes in @uc_write_list have the
+ * INTEL_UCLE_NOWR flag set (in which case, nothing is done
+ * to the file).
  *
  * @dirfd should be either an open directory for openat(), or
  * the special value AT_FDCWD.
@@ -1212,6 +1276,21 @@ static int write_intel_microcodes(const int dirfd,
 	assert(uc_write_list);
 	assert(filename);
 
+	/*
+	 * precalculate raw file size. It is less annoying than
+	 * having to seek back to write it later in cpio mode
+	 */
+	e = uc_write_list;
+	total_written = 0;
+	while (e && e->uc) {
+		if (likely(!(e->flags & INTEL_UCLE_NOWR)))
+			total_written += e->uc_size;
+		e = e->next;
+	}
+	if (unlikely(!total_written)) {
+		return ENODATA;	/* don't create empty files */
+	}
+
 	/* Unlink first */
 	if (unlink_files) {
 		if (unlinkat(dirfd, filename, 0) == -1) {
@@ -1235,17 +1314,6 @@ static int write_intel_microcodes(const int dirfd,
 		return err;
 	}
 
-	/*
-	 * precalculate raw file size. It is less annoying than
-	 * having to seek back to write it later in cpio mode
-	 */
-	e = uc_write_list;
-	total_written = 0;
-	while (e && e->uc) {
-		total_written += e->uc_size;
-		e = e->next;
-	}
-
 	if (ft == 1) {
 		rc = write_cpio_header(fd, total_written);
 		if (rc < 0)
@@ -1254,17 +1322,23 @@ static int write_intel_microcodes(const int dirfd,
 	}
 
 	while (uc_write_list && uc_write_list->uc) {
-		if (verbosity >= 3)
-			log_microcode_action("writing",
-					     filename, uc_write_list);
+		if (likely(!(uc_write_list->flags & INTEL_UCLE_NOWR))) {
+			if (verbosity >= 3)
+				log_microcode_action("writing",
+						     filename, uc_write_list);
 
-		rc = write_data(fd, uc_write_list->uc,
-				  uc_write_list->uc_size);
-		if (unlikely(rc < 0))
-			goto error;
+			rc = write_data(fd, uc_write_list->uc,
+					  uc_write_list->uc_size);
+			if (unlikely(rc < 0))
+				goto error;
 
+			entries_written++;
+		} else {
+			if (verbosity >= 3)
+				log_microcode_action("skipping",
+						     filename, uc_write_list);
+		}
 		uc_write_list = uc_write_list->next;
-		entries_written++;
 	}
 
 	if (ft == 1) {
@@ -1416,6 +1490,74 @@ static int is_selected(uint32_t cpuid, uint32_t pf_mask,
 	return filter_list_allow;
 }
 
+/* Microcode extended signature deduplication */
+
+static int xx_xtsdeduplist_add(struct intel_uclist_entry * const e, struct microcode_id_entry **list)
+{
+	struct microcode_id_entry *n;
+
+	n = malloc(sizeof(struct microcode_id_entry));
+	if (!n)
+		return ENOMEM;
+
+	n->id = e->uc;
+	n->next = *list;
+	*list = n;
+
+	return 0;
+}
+
+static int xx_xtsdeduplist_check_and_add(struct intel_uclist_entry * const e, struct microcode_id_entry **list)
+{
+	const void *id = e->uc;
+
+	while (*list) {
+		if ((*list)->id == id)
+			return EEXIST;
+		list = &((*list)->next);
+	}
+	return xx_xtsdeduplist_add(e, list);
+}
+
+static void xx_free_xtsdeduplist(struct microcode_id_entry *list)
+{
+	struct microcode_id_entry *e;
+
+	while (list) {
+		e = list;
+		list = list->next;
+		free(e);
+	}
+}
+
+static int uclist_annotate_extsig_dup(struct intel_uclist_entry * const uclist)
+{
+	struct microcode_id_entry *dlist = NULL;
+	struct intel_uclist_entry *e;
+	int rc = 0;
+
+	if (!extsig_tables_in_use)
+		return 0;
+
+	e = uclist;
+	while (e) {
+		if (unlikely(e->flags & INTEL_UCLE_HASXST)) {
+			rc = xx_xtsdeduplist_check_and_add(e, &dlist);
+			if (rc == EEXIST) {
+				e->flags |= INTEL_UCLE_NOWR;
+			} else if (!rc) {
+				e->flags &= ~INTEL_UCLE_NOWR;
+			} else {
+				break;
+			}
+		}
+		e = e->next;
+	}
+
+	xx_free_xtsdeduplist(dlist);
+	return (rc != EEXIST) ? rc : 0;
+}
+
 /* Microcode processing */
 
 static int xx_uclist_add_print_errors(int status)
@@ -1454,7 +1596,7 @@ static int xx_datefilter_loose_inplaceinsert(struct intel_uclist_entry **p,
 			e->flags |= INTEL_UCLE_SELECT;
 			rc = xx_uclist_add_print_errors(
 			        uclist_merge_signature(e->id, e->gid,
-				        e->cpuid, e->pf_mask,
+				        e->flags, e->cpuid, e->pf_mask,
 					e->uc_rev, e->uc, e->uc_size,
 					allow_downgrade, p));
 			if (rc)
@@ -1479,7 +1621,7 @@ static int xx_process_ucode_signature_cb(void *userdata,
 	struct intel_ucode_metadata m;
 	intel_ucode_status_t s;
 	int add_status;
-	uint32_t uce_flags;
+	uint32_t uce_flags = 0;
 
 	assert(ctx);
 
@@ -1492,7 +1634,12 @@ static int xx_process_ucode_signature_cb(void *userdata,
 		return 1;
 	}
 
-	uce_flags = (sig_count)? INTEL_UCLE_EXTSIG : 0;
+	if (m.extsig_count) {
+		uce_flags = (sig_count) ?
+			(INTEL_UCLE_HASXST | INTEL_UCLE_EXTSIG) :
+			INTEL_UCLE_HASXST;
+		extsig_tables_in_use = 1;
+	}
 	if (is_selected(cpuid, pf_mask, uc_filter_list))
 		uce_flags |= (is_in_date_range(&m)) ?
 				INTEL_UCLE_SELECT : INTEL_UCLE_SELID;
@@ -1536,7 +1683,7 @@ static int xx_process_ucode_signature_cb(void *userdata,
 
 	if (uce_flags & INTEL_UCLE_SELECT) {
 		add_status = uclist_merge_signature(ctx->current_uc,
-					ctx->current_bundle->id,
+					ctx->current_bundle->id, uce_flags,
 					cpuid, pf_mask, m.revision,
 					uc, uc_size, allow_downgrade,
 					&microcodes);
@@ -1674,11 +1821,18 @@ static int do_process_microcodes(void)
 
 static int do_write_microcode(const char * const filename, int ft)
 {
+	int rc;
+
 	if (!microcodes)
 		return 0;
 
 	print_msg(1, "Writing selected microcodes to: %s", filename);
-	return write_intel_microcodes(AT_FDCWD, filename, ft, microcodes);
+	rc = write_intel_microcodes(AT_FDCWD, filename, ft, microcodes);
+	if (rc == ENODATA) {
+		print_msg(1, "All microcodes in %s were skipped, file unchanged", filename);
+		return 0;
+	}
+	return rc;
 }
 
 static int do_upload_microcode(const char * const filename)
@@ -1736,6 +1890,7 @@ static int do_write_named(const char * const dirname)
 
 		memcpy(&e, p, sizeof(e));
 		e.next = NULL;
+		e.flags &= ~INTEL_UCLE_NOWR;
 		rc = write_intel_microcodes(dirfd, fn, 0, &e);
 		if (!rc)
 			count++;
@@ -1804,7 +1959,7 @@ static int do_write_firmware(const char * const dirname)
 		cpuid = p->cpuid;
 		while (p && p->cpuid == cpuid) {
 			add_status = uclist_merge_signature(p->id, p->gid,
-						p->cpuid, p->pf_mask,
+						p->flags, p->cpuid, p->pf_mask,
 						p->uc_rev, p->uc, p->uc_size,
 						0, &samecpuid_list);
 			if (xx_uclist_add_print_errors(add_status)) {
@@ -1828,6 +1983,10 @@ static int do_write_firmware(const char * const dirname)
 		snprintf(fn, sizeof(fn), "%02x-%02x-%02x",
 			     x86_family, x86_model, x86_mask);
 
+		rc = uclist_annotate_extsig_dup(samecpuid_list);
+		if (rc)
+			goto err_exit;
+
 		rc = write_intel_microcodes(dirfd, fn, 0, samecpuid_list);
 
 		free_uclist(samecpuid_list);
@@ -1835,6 +1994,8 @@ static int do_write_firmware(const char * const dirname)
 
 		if (!rc)
 			count++;
+		if (rc == ENODATA)
+			rc = 0;
 	}
 	if (count)
 		print_msg(2, "%lu file(s) were written into %s", count, dirname);
@@ -1853,15 +2014,12 @@ static int scan_system_processors(void)
 	uint32_t cpuid_buf[8];
 	struct microcode_filter_entry *uc_cpu = NULL;
 
-	int cpuid_fd = -1;
+	int cpuid_fd;
 	unsigned int i = 0;
 	unsigned int ncpu = 0;
 	int rc = 0;
 
 	while (1) {
-		if (cpuid_fd != -1)
-			close(cpuid_fd);
-
 		snprintf(cpuid_device, sizeof(cpuid_device),
 			 CPUID_DEVICE_BASE, i);
 		cpuid_fd = open(cpuid_device, O_RDONLY);
@@ -1895,12 +2053,12 @@ static int scan_system_processors(void)
 			print_err("%s: access to CPUID(0) and CPUID(1) failed",
 				  cpuid_device);
 			/* this is in the should not happen list, so abort */
+			close(cpuid_fd);
 			rc = 1;
 			goto err_out;
 		}
 
 		close(cpuid_fd);
-		cpuid_fd = -1;
 		ncpu++;
 
 		/* Is it a supported Intel processor ? */
@@ -1966,9 +2124,6 @@ static int scan_system_processors(void)
 	print_msg(2, "checked the signature of %u processor(s)", ncpu);
 
 err_out:
-	if (cpuid_fd != -1)
-		close(cpuid_fd);
-
 	if (uc_cpu)
 		free_filter_list(uc_cpu);
 
@@ -2285,6 +2440,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 	switch (key) {
 	case 'h':
 		argp_state_help(state, stdout, ARGP_HELP_STD_HELP);
+		break; /* usually not reached */
 
 	case 'q':
 		verbosity = 0;
@@ -2301,7 +2457,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 		break;
 
 	case 't':
-		if (!arg || strlen(arg) > 1)
+		if (strlen(arg) > 1)
 			argp_error(state, "unknown file type: %s\n", arg);
 		switch (*arg) {
 		case 'd': /* .dat */
@@ -2383,6 +2539,7 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 			break; /* success */
 		case EINVAL:
 			argp_error(state, "invalid filter: %s", arg);
+			break; /* not reached */
 		default:
 			argp_failure(state, EXIT_SWFAILURE, rc,
 				     "could not add filter '%s'", arg);
@@ -2495,8 +2652,8 @@ int main(int argc, char *argv[])
 
 	if (command_line_actions & IUCODE_DOMASK_NEEDSUC) {
 		if (microcodes) {
-			rc = 0;
-			if (upload_microcode)
+			rc = uclist_annotate_extsig_dup(microcodes);
+			if (!rc && upload_microcode)
 				rc = do_upload_microcode(upload_microcode);
 			if (!rc && write_microcode)
 				rc = do_write_microcode(write_microcode, 0);
